@@ -6,6 +6,13 @@ from .zip import zip_directory
 from .zip import unzip
 from .get_creation_date import get_creation_date
 from .get_creation_date import get_modification_date
+from .exceptions import DiskError
+from .exceptions import RenameError
+from .exceptions import PathDoesNotExistError
+from .exceptions import PathExistsError
+from .exceptions import NotAFileError
+from .exceptions import DirectoryNotFoundError
+
 
 import re
 from zipfile import ZIP_DEFLATED
@@ -14,6 +21,7 @@ from pathlib import Path as _Path
 import warnings
 import os
 import shutil
+from filecmp import cmp as compare_files
 
 
 def get_parent_directory(path):
@@ -39,39 +47,112 @@ class Path:
 		:type new_name: str or Path
 		:rtype: Path
 		"""
-		if isinstance(new_name, Path):
-			new_name = new_name.name_and_extension
-
-		shutil.move(self.absolute_path, (self.parent_directory + new_name).absolute_path)
-		self._string = self._string[:-len(self.name_and_extension)] + new_name
-		return self
+		return self.move_and_rename(new_name=new_name, new_directory=self.parent_directory)
 
 	def move(self, new_directory):
 		"""
 		:type new_directory: str or Path
 		:rtype: Path
 		"""
-		new_directory = Path(new_directory)
-		if not new_directory.exists():
-			new_directory.make_dir()
-		new_path = (new_directory + self.name_and_extension)
-		shutil.move(self.absolute_path, new_path.absolute_path)
-		self._string = new_path.absolute_path
-		return self
+		return self.move_and_rename(new_name=self.name_and_extension, new_directory=new_directory)
 
-	def copy(self, new_path):
+	def rename_and_move(self, **kwargs):
+		return self.move_and_rename(**kwargs)
+
+	def move_and_rename(self, new_name=None, new_directory=None, new_path=None):
 		"""
-		:type new_path str or Path
+		:type new_name: str or Path
+		:type new_directory: str or Path
+		:type new_path: str or Path
 		:rtype: Path
 		"""
 		if not self.exists():
-			raise FileExistsError(f'{self} does not exist!')
+			raise PathDoesNotExistError(f'"{self.absolute}" does not exist to be renamed!')
 
-		new_path = Path(new_path)
-		if new_path.exists() and new_path.is_directory() and self.is_file():
-			new_path = new_path + self.name_and_extension
+		if new_name is None and new_directory is None and new_path is None:
+			raise RenameError('at least one of new_name, new_directory, new_path should be provided!')
+
+		if new_path is not None and (new_name is not None or new_directory is not None):
+			raise RenameError('providing both new_path and either of new_directory or new_name is ambiguous!')
+
+		if new_path is not None:
+			new_path = Path(new_path)
+			new_directory = new_path.parent_directory
+
+		else:
+			if new_name is None:
+				new_name = self.name_and_extension
+			elif isinstance(new_name, Path):
+				new_name = new_name.name_and_extension
+
+			if new_directory is None:
+				new_directory = self.parent_directory
+
+			new_directory = Path(new_directory)
+			new_path = new_directory + new_name
+
+		if not new_directory.exists():
+			new_directory.make_dir()
+
+		old_path = str(self.absolute)
+
+		if new_path.exists():
+			raise PathExistsError(f'"{old_path}" cannot be renamed to "{new_path.absolute}" because it already exists!')
+
+		shutil.move(self.absolute_path, new_path.absolute_path)
+		self._string = new_path.absolute_path
+		if not self.exists():
+			raise RenameError(f'could not rename "{old_path}" to "{self.absolute}"')
+		return self
+
+	def copy(self, new_path=None, new_directory=None, clean_copy=True, echo=0):
+		"""
+		:type new_path str or Path
+		:type echo: int or bool
+		:param bool clean_copy: if True, first delete destination
+		:rtype: Path
+		"""
+		if not self.exists():
+			raise FileNotFoundError(self)
+
+		if new_path is None and new_directory is not None:
+			new_directory = Path(new_directory)
+			if not new_directory.exists():
+				raise DirectoryNotFoundError(new_directory)
+			new_path = new_directory + self.name_and_extension
+
+		elif new_path is not None and new_directory is None:
+			new_path = Path(new_path)
+
+		elif new_path is None and new_directory is None:
+			raise TypeError('new_path and new_directory are both None!')
+
+		elif new_path is not None and new_directory is not None:
+			raise TypeError('Please make up your mind and provide only one of new_path or new_directory')
+
+		if self.is_file():
+			if new_path.exists() and new_path.is_directory():
+				raise IsADirectoryError(new_path)
+			if echo:
+				print(f'Copying "{self.absolute_path}" to "{new_path.absolute_path}"')
+			if clean_copy and new_path.exists():
+				new_path.delete(echo=echo)
+			shutil.copy2(self.path, new_path.path)
+
+		elif self.is_directory():
+			if new_path.exists() and new_path.is_file():
+				raise NotADirectoryError(new_path)
+			if echo:
+				print(f'Copying "{self.absolute_path}" to "{new_path.absolute_path}"')
+			if not new_path.exists():
+				new_path.make_dir()
+			elif clean_copy:
+				new_path.delete()
+				new_path.make_dir()
+
+			for path in self.list():
+				path.copy(new_directory=new_path, clean_copy=clean_copy)
 
-		shutil.copy2(self.path, new_path.path)
 		return new_path
 
 	def __str__(self):
@@ -327,7 +408,7 @@ class Path:
 		else:
 			return 'directory'
 
-	def __add__(self, other):
+	def __truediv__(self, other):
 		"""
 		:type other: Path or str
 		:rtype: Path
@@ -345,6 +426,8 @@ class Path:
 			show_size=self._show_size
 		)
 
+	__add__ = __truediv__
+
 	def _sort_key(self):
 		return self.type, self.get_absolute()
 
@@ -434,26 +517,30 @@ class Path:
 	def get(self, full_name):
 		return [x for x in self.list() if x.full_name == full_name][0]
 
-	def make_directory(self, name=None, ignore_if_exists=True):
+	def make_directory(self, name=None, ignore_if_exists=True, echo=0):
 		if name:
 			path = self+name
 		else:
 			path = self
 
+		if echo:
+			print(f'Making directory "{path.absolute_path}"')
 		make_dir(path=path.path, ignore_if_exists=ignore_if_exists)
 
 		return path
 
-	def make_parent_directory(self, ignore_if_exists=True):
+	def make_parent_directory(self, ignore_if_exists=True, echo=0):
 		if self.parent_directory:
-			self.parent_directory.make_directory(ignore_if_exists=ignore_if_exists)
+			self.parent_directory.make_directory(ignore_if_exists=ignore_if_exists, echo=echo)
 		return self.parent_directory
 
-	def delete(self, name=None):
+	def delete(self, name=None, echo=0):
 		if name:
 			to_delete = self + name
 		else:
 			to_delete = self
+		if echo:
+			print(f'Deleting "{to_delete.absolute_path}"')
 		delete(path=to_delete.path)
 
 	def delete_directory(self, name):
@@ -575,7 +662,177 @@ class Path:
 			os.chdir(working_directory)
 			return result
 
+	def is_inside(self, path):
+		"""
+		checks if self is inside another path
+		:type path: Path
+		:rtype: bool
+		"""
+		if not path.is_directory():
+			raise NotADirectoryError(path)
+		return self.absolute_path.startswith(path.absolute_path)
+
+	def contains(self, path):
+		"""
+		checks if self contains another path
+		:type path: Path
+		:rtype: bool
+		"""
+		return path.is_inside(path=self)
+
+	def is_the_same_file(self, other):
+		"""
+		checks if self is the same file as path
+		:type other: Path
+		:rtype: bool
+		"""
+		if not self.exists():
+			FileNotFoundError(self)
+
+		if not other.exists():
+			FileNotFoundError(other)
+
+		if not self.is_file():
+			raise NotAFileError(self)
+
+		if not other.is_file():
+			raise NotAFileError(other)
+
+		return compare_files(self.absolute_path, other.absolute_path)
+
+	def is_the_same_directory(self, other):
+		"""
+		checks if self is the same directory as path
+		:type other: Path
+		:rtype: bool
+		"""
+		if not self.exists():
+			DirectoryNotFoundError(self)
+
+		if not other.exists():
+			DirectoryNotFoundError(other)
+
+		if not self.is_directory():
+			raise NotADirectoryError(self)
+
+		if not other.is_directory():
+			raise NotADirectoryError(other)
+
+		self_file_names = {f.name_and_extension: f for f in self.files}
+		other_file_names = {f.name_and_extension: f for f in other.files}
+
+		for name, file in self_file_names.items():
+			if name not in other_file_names:
+				return False
+			if not file.is_the_same_file(other_file_names[name]):
+				return False
+
+		for name in other_file_names.keys():
+			if name not in self_file_names:
+				return False
+
+		self_dir_names = {d.name_and_extension: d for d in self.directories}
+		other_dir_names = {d.name_and_extension: d for d in other.directories}
+
+		for name, directory in self_dir_names.items():
+			if name not in other_dir_names:
+				return False
+
+		for name in other_dir_names.keys():
+			if name not in self_dir_names:
+				return False
+
+		for name, directory in self_dir_names.items():
+			other_directory = other_dir_names[name]
+
+			if not directory.is_the_same_directory(other_directory):
+				return False
+
+		return True
+
+	def mimic(self, other, ignore_function=None, echo=1):
+		"""
+		syncs self with other path which must be a directory by mimicking it
+		:type other: Path
+		:type ignore_function: callable or NoneType
+		:rtype: Path
+		"""
+		if echo:
+			print(f'"{self.absolute_path}" mimicking "{other.absolute_path}" ')
+		if not self.exists():
+			self.make_directory()
+		if not other.exists():
+			raise PathDoesNotExistError(other)
+		if other.is_file():
+			raise NotADirectoryError(other)
+		if self.is_inside(path=other):
+			raise DiskError(f'"{self.path}" is inside "{other.path}"!')
+		if self.contains(path=other):
+			raise DiskError(f'"{self.path}" contains "{other.path}"!')
+
+		self_files = {file.name_and_extension: file for file in self.files}
+		other_files = {file.name_and_extension: file for file in other.files}
+
+		if ignore_function is None:
+			def ignore_function(x):
+				return False
+
+		for name, file in self_files.items():
+			if ignore_function(file):
+				continue
+
+			elif name not in other_files:
+				file.delete(echo=echo)
+
+		for name, other_file in other_files.items():
+			if ignore_function(other_file):
+				continue
+
+			elif name not in self_files:
+				self_file = self + name
+				if ignore_function(self_file):
+					continue
+				else:
+					other_file.copy(new_path=self_file.absolute_path, echo=echo)
+
+			else:
+				self_file = self_files[name]
+				if ignore_function(self_file):
+					continue
+
+				elif not other_file.is_the_same_file(self_file):
+					self_file.delete(echo=echo)
+					other_file.copy(new_path=self_file.absolute_path, echo=echo)
+
+		self_directories = {d.name_and_extension: d for d in self.directories}
+		other_directories = {d.name_and_extension: d for d in other.directories}
+
+		for name, directory in self_directories.items():
+			if ignore_function(directory):
+				continue
+			elif name not in other_directories:
+				directory.delete(echo=echo)
+
+		for name, other_directory in other_directories.items():
+			if ignore_function(other_directory):
+				continue
+
+			elif name not in self_directories:
+				self_directory = self + name
+				if ignore_function is not None and ignore_function(self_directory):
+					continue
+
+				else:
+					other_directory.copy(new_path=self_directory.absolute_path, echo=echo)
+			else:
+				self_directory = self_directories[name]
+				if ignore_function(self_directory):
+					continue
+				else:
+					self_directory.mimic(other_directory, ignore_function=ignore_function, echo=echo)
+
 	# aliases
+
 	ls = list
 	dir = list
 	md = make_directory
@@ -584,3 +841,8 @@ class Path:
 	delete_dir = delete_directory
 	pickle = save
 	unpickle = load
+	replicate = mimic
+	is_same_file = is_the_same_file
+	is_same_directory = is_the_same_directory
+	is_the_same_dir = is_the_same_directory
+	is_same_dir = is_same_directory
